from inspect import Signature
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydantic import validate_arguments

from starlite.exceptions import ImproperlyConfiguredException
from starlite.handlers.base import BaseRouteHandler
from starlite.types import AsyncAnyCallable, ExceptionHandlersMap, Guard
from starlite.utils import Ref, is_async_callable

if TYPE_CHECKING:
    from starlite.types import MaybePartial  # nopycln: import # noqa: F401


class ASGIRouteHandler(BaseRouteHandler["ASGIRouteHandler"]):
    """ASGI Route Handler decorator.

    Use this decorator to decorate ASGI applications.
    """

    __slots__ = ("is_mount", "is_static")

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        path: Union[Optional[str], Optional[List[str]]] = None,
        *,
        exception_handlers: Optional[ExceptionHandlersMap] = None,
        guards: Optional[List[Guard]] = None,
        name: Optional[str] = None,
        opt: Optional[Dict[str, Any]] = None,
        is_mount: bool = False,
        is_static: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize ``ASGIRouteHandler``.

        Args:
            exception_handlers: A dictionary that maps handler functions to status codes and/or exception types.
            guards: A list of :class:`Guard <starlite.types.Guard>` callables.
            name: A string identifying the route handler.
            opt: A string key dictionary of arbitrary values that can be accessed in :class:`Guards <starlite.types.Guard>` or
                wherever you have access to :class:`Request <starlite.connection.request.Request>` or :class:`ASGI Scope <starlite.types.Scope>`.
            path: A path fragment for the route handler function or a list of path fragments. If not given defaults to '/'
            is_mount: A boolean dictating whether the handler's paths should be regarded as mount paths. Mount path accept
                any arbitrary paths that begin with the defined prefixed path. For example, a mount with the path `/some-path/`
                will accept requests for `/some-path/` and any sub path under this, e.g. `/some-path/sub-path/` etc.
            is_static: A boolean dictating whether the handler's paths should be regarded as static paths. Static paths
                are used to deliver static files.
            **kwargs: Any additional kwarg - will be set in the opt dictionary.
        """
        self.is_mount = is_mount or is_static
        self.is_static = is_static
        super().__init__(path, exception_handlers=exception_handlers, guards=guards, name=name, opt=opt, **kwargs)

    def __call__(self, fn: "AsyncAnyCallable") -> "ASGIRouteHandler":
        """Replace a function with itself."""
        self.fn = Ref["MaybePartial[AsyncAnyCallable]"](fn)
        self.signature = Signature.from_callable(fn)
        self._validate_handler_function()
        return self

    def _validate_handler_function(self) -> None:
        """Validate the route handler function once it's set by inspecting its return annotations."""
        super()._validate_handler_function()

        if self.signature.return_annotation not in {None, "None"}:
            raise ImproperlyConfiguredException("ASGI handler functions should return 'None'")

        if any(key not in self.signature.parameters for key in ("scope", "send", "receive")):
            raise ImproperlyConfiguredException(
                "ASGI handler functions should define 'scope', 'send' and 'receive' arguments"
            )
        if not is_async_callable(self.fn.value):
            raise ImproperlyConfiguredException("Functions decorated with 'asgi' must be async functions")


asgi = ASGIRouteHandler



from typing import Any, Optional

from pydantic import BaseModel

from starlite.exceptions import MissingDependencyException

try:
    from redis.asyncio import Redis
    from redis.asyncio.connection import ConnectionPool
except ImportError as e:
    raise MissingDependencyException(
        "To use starlite.redis_cache_backend, install starlite with 'redis' extra, e.g. `pip install starlite[redis]`"
    ) from e


from starlite.cache.base import CacheBackendProtocol


class RedisCacheBackendConfig(BaseModel):
    """Redis cache backend configuration."""

    url: str
    """Redis connection URL."""
    db: Optional[int] = None
    """Redis DB ID (optional)"""
    port: Optional[int] = None
    """Redis port (optional)"""
    username: Optional[str] = None
    """A username to use when connecting to Redis (optional)"""
    password: Optional[str] = None
    """A password to use when connecting to Redis (optional)"""


class RedisCacheBackend(CacheBackendProtocol):
    """Redis-based cache backend."""

    def __init__(self, config: RedisCacheBackendConfig):
        """Initialize ``RedisCacheBackend``

        Args:
            config: required configuration to connect to Redis.
        """
        self._config = config
        self._redis_int: Redis = None  # type: ignore[assignment]

    @property
    def _redis(self) -> Redis:
        if not self._redis_int:
            pool = ConnectionPool.from_url(**self._config.dict(exclude_unset=True))
            self._redis_int = Redis(connection_pool=pool)

        return self._redis_int

    async def get(self, key: str) -> Any:  # pylint: disable=invalid-overridden-method
        """Retrieve a value from cache corresponding to the given key.

        Args:
            key: name of cached value.

        Returns:
            Cached value if existing else ``None``.
        """

        return await self._redis.get(key)

    async def set(self, key: str, value: Any, expiration: int) -> None:  # pylint: disable=invalid-overridden-method
        """Set a value in cache for a given key for a duration determined by expiration.

        Args:
            key: key to cache ``value`` under.
            value: the value to be cached.
            expiration: expiration of cached value in seconds.

        Notes:
            - expiration is in seconds.
            - return value is not used by Starlite internally.

        Returns:
            None
        """

        await self._redis.set(key, value, ex=expiration)

    async def delete(self, key: str) -> None:  # pylint: disable=invalid-overridden-method
        """Delete a value from the cache and removes the given key.

        Args:
            key: key to be deleted from the cache.

        Notes:
            - return value is not used by Starlite internally.

        Returns:
            None
        """

        await self._redis.delete(key)



import re
from collections import deque
from typing import (
    Any,
    Deque,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TypeGuard, get_args

from starlite.types.builtin_types import NoneType
from starlite.utils.predicates import is_class_and_subclass

T = TypeVar("T")
UnionT = TypeVar("UnionT", bound="Union")

tuple_types_regex = re.compile(
    "^"
    + "|".join(
        [*[repr(x) for x in (List, Sequence, Iterable, Iterator, Tuple, Deque)], "tuple", "list", "collections.deque"]
    )
)


def annotation_is_iterable_of_type(
    annotation: Any,
    type_value: Type[T],
) -> "TypeGuard[Iterable[T]]":
    """Determine if a given annotation is an iterable of the given type_value.

    Args:
        annotation: A type annotation.
        type_value: A type value.

    Returns:
        A type-guard boolean.
    """
    if (args := get_args(annotation)) and (
        isinstance(annotation, (List, Sequence, Iterable, Iterator, Tuple, Deque, tuple, list, deque))  # type: ignore
        or tuple_types_regex.match(repr(annotation))
    ):
        return args[0] is type_value or isinstance(args[0], type_value) or is_class_and_subclass(args[0], type_value)
    return False


def make_non_optional_union(annotation: Optional[UnionT]) -> UnionT:
    """Make a :data:`Union <typing.Union>` type that excludes ``NoneType``.

    Args:
        annotation: A type annotation.

    Returns:
        The union with all original members, except ``NoneType``.
    """
    args = tuple(tp for tp in get_args(annotation) if tp is not NoneType)
    return cast("UnionT", Union[args])  # pyright: ignore



from functools import lru_cache
from http.cookies import _unquote as unquote_cookie
from typing import Any, Dict, Iterable, List, Tuple, Union
from urllib.parse import unquote

from fast_query_parsers import parse_query_string as fast_parse_query_string
from fast_query_parsers import parse_url_encoded_dict


@lru_cache(1024)
def parse_url_encoded_form_data(encoded_data: bytes) -> Dict[str, Any]:
    """Parse an url encoded form data dict.

    Args:
        encoded_data: The encoded byte string.

    Returns:
        A parsed dict.
    """
    return parse_url_encoded_dict(encoded_data)


@lru_cache(1024)
def parse_query_string(query_string: bytes) -> Tuple[Tuple[str, Any], ...]:
    """Parse a query string into a tuple of key value pairs.

    Args:
        query_string: A query string.
        encoding: The encoding to use.

    Returns:
        A tuple of key value pairs.
    """
    _bools = {"true": True, "false": False, "True": True, "False": False}
    return tuple((k, v if v not in _bools else _bools[v]) for k, v in fast_parse_query_string(query_string, "&"))


@lru_cache(1024)
def parse_cookie_string(cookie_string: str) -> Dict[str, str]:
    """Parse a cookie string into a dictionary of values.

    Args:
        cookie_string: A cookie string.

    Returns:
        A string keyed dictionary of values
    """
    output: Dict[str, str] = {}
    cookies = [cookie.split("=", 1) if "=" in cookie else ("", cookie) for cookie in cookie_string.split(";")]
    for k, v in filter(lambda x: x[0] or x[1], ((k.strip(), v.strip()) for k, v in cookies)):
        output[k] = unquote(unquote_cookie(v))
    return output


@lru_cache(1024)
def _parse_headers(headers: Tuple[Tuple[bytes, bytes], ...]) -> Dict[str, str]:
    """Parse ASGI headers into a dict of string keys and values.

    Args:
        headers: A tuple of bytes two tuples.

    Returns:
        A string / string dict.
    """
    return {k.decode(): v.decode() for k, v in headers}


def parse_headers(headers: Iterable[Union[Tuple[bytes, bytes], List[bytes]]]) -> Dict[str, str]:
    """Parse ASGI headers into a dict of string keys and values.

    Since the ASGI protocol only allows for lists (not tuples) which cannot be hashed,
    this function will convert the headers to a tuple of tuples before invoking the cached function.
    """
    return _parse_headers(tuple(tuple(h) for h in headers))



from starlite.datastructures.background_tasks import BackgroundTask, BackgroundTasks
from starlite.datastructures.cookie import Cookie
from starlite.datastructures.headers import (
    CacheControlHeader,
    ETag,
    Headers,
    MutableScopeHeaders,
)
from starlite.datastructures.multi_dicts import (
    FormMultiDict,
    ImmutableMultiDict,
    MultiDict,
)
from starlite.datastructures.pagination import (
    AbstractAsyncClassicPaginator,
    AbstractAsyncCursorPaginator,
    AbstractAsyncOffsetPaginator,
    AbstractSyncClassicPaginator,
    AbstractSyncCursorPaginator,
    AbstractSyncOffsetPaginator,
    ClassicPagination,
    CursorPagination,
    OffsetPagination,
)
from starlite.datastructures.provide import Provide
from starlite.datastructures.response_containers import (
    File,
    Redirect,
    ResponseContainer,
    Stream,
    Template,
)
from starlite.datastructures.response_header import ResponseHeader
from starlite.datastructures.state import ImmutableState, State
from starlite.datastructures.upload_file import UploadFile
from starlite.datastructures.url import URL, Address, make_absolute_url

__all__ = (
    "AbstractAsyncClassicPaginator",
    "AbstractAsyncCursorPaginator",
    "AbstractAsyncOffsetPaginator",
    "AbstractSyncClassicPaginator",
    "AbstractSyncCursorPaginator",
    "AbstractSyncOffsetPaginator",
    "Address",
    "BackgroundTask",
    "BackgroundTasks",
    "CacheControlHeader",
    "ClassicPagination",
    "Cookie",
    "CursorPagination",
    "ETag",
    "File",
    "FormMultiDict",
    "Headers",
    "ImmutableMultiDict",
    "ImmutableState",
    "MultiDict",
    "MutableScopeHeaders",
    "OffsetPagination",
    "Provide",
    "Redirect",
    "ResponseContainer",
    "ResponseHeader",
    "State",
    "Stream",
    "Template",
    "URL",
    "UploadFile",
    "make_absolute_url",
)
